#
# Virtual Traces Experiment Software Distribution.
# This code is licensed under GPL v2, and is copyright (C) Jack Whitham 2008.
# See PATCH_NOTE.
#

import os, sys, capture_metas, pickle, shutil

# Things that are needed for measurement: always on
#DP_IGNORE_CACHE_MISSES=1
#DP_SHORT_TLB=1
# Things that are needed for VTs, which might possibly reduce throughput
#DP_RESYNC=1
#DP_IO_BRANCHES=1
#DP_ALLOW_LOAD_PASSING=1
#DP_STATIC_DISAMBIG=1
#DP_NO_FORWARD=1
# Things that are useful for debugging
#DP_PARANOID=1

M5VER = "ARM_SE/m5.debug"
BENCH_PATH = "../benchmarks/bin"

def Crc(rc, name):
    if ( rc == 0 ): 
        return
    print '%s rc = %d' % (name, rc)
    sys.exit(1)

def Run_VT(name, capture_timings, execute_program):
    vt_parameters = {
        'cmd' : '%s.elf' % name,
        'meta_in' : '%s_meta.txt' % name,
        'meta_out' : '%s_meta.out' % name,
        'max_length' : capture_metas.MAX_META_INST_LENGTH + 1,
        'warnings_out' : '%s_warn.txt' % name,
        'start_pc' : 0x80b0,
        'capture_timings' : capture_timings,
        'runtime' : 0,
        }

    if ( not capture_timings ):
        vt_parameters[ 'meta_in' ] = vt_parameters[ 'meta_out' ]
    if ( execute_program ):
        vt_parameters[ 'runtime' ] = int(1e10)

    pickle.dump(vt_parameters, file('%s.dat' % name, 'wb'))

    flags = "Fetch,Commit"
    #flags = "O3CPUAll"
    
    return os.system("build/%s --trace-flags=%s "
            "configs/example/vt1.py --caches --detailed -c %s.dat "
            ">%s_1.txt 2>%s_2.txt" % (M5VER, flags, name, name, name))

def Check_Path(name):
    print 'Split', name
    Crc(os.system('csplit -f %s_run_ %s_1.txt '
                '"/^XXX Going to execution mode/"' % (name, name)),
                "split execution")

    print 'Results for', name
    Crc(os.system(r"sed -ne 's/^.*Committed pc \(.*\) .sn:.*$/\1/p' "
                "<%s_run_01 >%s_co3.txt" % (name, name)), "commits o3")

    diff_rc = os.system("diff %s_ccm.txt %s_co3.txt > %s_diff.txt" %
                    (name, name, name))

    lines = len([ x for x in file('%s_diff.txt' % name) ])
    rc = ( lines <= 3 )
    if ( not rc ):
        print 'Danger! diff contains %u lines - early exit from execution?' % (
                    lines)

    return rc

def Find_Main(name):
    Crc(os.system("../findmain %s.elf wcetmain >%s_loc.txt" % (name, name)), 
            "fm")
    Crc(os.system("../findmain %s.elf main >>%s_loc.txt" % (name, name)), 
            "fm")
    fields = file("%s_loc.txt" % name).read().split()
    fields.pop(3)
    fields.pop(0)
    [ wcetmain_start , wcetmain_size ,
        main_start , main_size ] = [ int(x, 16) for x in fields ]
    wcetmain_end = wcetmain_start + wcetmain_size
    main_end = main_start + main_size
    return (wcetmain_start, wcetmain_end, main_start, main_end)


def Make_Timings(name, vt_mode):
    print 'Run (tlog)', name
    Crc(os.system("build/%s --trace-flags=Fetch,Exec "
        "configs/example/se.py -c %s.elf >%s_run.txt" % (M5VER,
                name, name)), "tlog")


    (wcetmain_start, wcetmain_end, main_start, main_end) = Find_Main(name)
    count = 0
    in_benchmark = 0

    ccm_file = file('%s_ccm.txt' % name, 'wt')
    pc = 0
    for line in file('%s_run.txt' % name):
        fields = line.split(':')
        if ( len(fields) < 3 ):
            continue
        if ( fields[ 2 ] == ' Fetch' ):
            # Committed this instruction:
            # (Fetch == commit for simple CPU)
            pc = int(fields[ 4 ].split()[ 0 ],16)
            ccm_file.write('0x%x\n' % pc)

            if (( in_benchmark == 0 )
            and ( wcetmain_start <= pc < wcetmain_end )):
                in_benchmark = 1
            elif (( in_benchmark == 1 )
            and ( main_start <= pc < main_end )):
                in_benchmark = 2
            
            if ( in_benchmark == 1 ):
                count += 1

    assert in_benchmark == 2

    ccm_file.close()

    output = "%s_meta.txt" % name
    if ( not vt_mode ):
        output = "%s_meta.out" % name

    tlog_input = "%s_run.txt" % name

    capture_metas.Process(tlog_input=tlog_input,
                output=output,
                path_output='%s_path.txt' % name)

    print 'Run (timing)', name, vt_mode
    main_rc = Run_VT(name=name, capture_timings=vt_mode, execute_program=True)
    os.system("cp %s_1.txt %s_3.txt" % (name, name))

    cp_ok = Check_Path(name) 
    ok = cp_ok and ( main_rc == 0 )

    Crc(os.system("tail -n 100 %s_run_01 > %s_result.txt" % (name, name)), 
                "tail") 

    MAGIC_NUMBER = "magic number"
    timing = 0
    for line in file('%s_result.txt' % name):
        if ( line.startswith(MAGIC_NUMBER) ):
            timing = int(line[ len(MAGIC_NUMBER) : ])

    # Compare O3VTC trace timings with IIO
    meta_inst_exits = capture_metas.Compare_IIO_O3VTC(tlog_input, 
            "%s_meta.out" % name, "%s_cmp.txt" % name)

    # Also determine mispredict count
    # and trace start count
    out = Scan_Run_Log('%s_run_01' % name, meta_inst_exits)

    if ( vt_mode and not ok ):
        print 'Make_Timings - Not OK'
        print 'Program', name
        print 'Check_Path', cp_ok
        print 'main_rc', main_rc
        return None

    Cleanup(name)
    out.update({"inst_count" : count, "timing" : timing})
    return out

def Get_Parameter(line, name, base):
    i = line.find(name)
    assert i >= 0
    return int(line[ i + len(name) : ].split()[ 0 ], base)

def Scan_Run_Log(name, meta_inst_exits):

    exit_count = 0
    trace_count = 0
    total_o3vtc_cost = 0
    total_iio_cost = 0
    be_count = 0
    be_cost = 0

    nvt_pc = vt_pc = 0
    in_benchmark = 0
    mi_exits = None
    valid = False
    valid_next = False
    X1_BS = "X1 BS Enter "
    predictions = []
    #print 'SRL START'

    for line in file(name):
        #print line.rstrip()
        x1bs = line.find(X1_BS)
        if ( x1bs > 0 ):
            x1bs += len(X1_BS)
            new_state = line[ x1bs : ].split()[ 0 ]
            if ( new_state == "Timing" ):
                in_benchmark = 1
                trace_count += 1
                valid = True
                valid_next = True
                #print 'VALIDATE'
            elif ( new_state == "Done" ) :
                in_benchmark = 2
                valid_next = False
                #print 'DONE'

        et = '?'
        exit = None

        if ( line.find("global: X1 Issuing MI. TS=Fetching") > 0 ):
            # Start of a trace
            if ( in_benchmark == 1 ):
                trace_count += 1

            # End of previous trace?
            if ( mi_exits != None ):
                exit = mi_exits[ -1 ]

            # What vt is this?
            nvt_pc = Get_Parameter(line, "pc=", 16)
            predictions = []
            mi_exits = meta_inst_exits[ nvt_pc ]
            #print 'ISSUE %x' % nvt_pc

            if ( exit != None ):
                #print 'MAIN EXIT'
                et = 'main'

        elif ( line.find("Squashing, setting PC to:") > 0 ):
            # Side exit being taken
            if ( in_benchmark == 1 ):
                exit_count += 1
            
        elif ( line.find("Mispredict TPID received") > 0 ):
            # Side exit being taken *in a trace*
            tpid = Get_Parameter(line, "received:", 10)

            side_exit_number = None
            for (i, p) in enumerate(predictions):
                if ( p == tpid ):
                    side_exit_number = i
                    break

            if ( side_exit_number != None ):
                exit = mi_exits[ side_exit_number ]
                #print 'SIDE EXIT %u' % side_exit_number
                et = 'side %u' % side_exit_number
                assert exit != None

            mi_exits = None

        elif ( line.find("Trace overrides normal prediction for") > 0 ):
            # Register issued prediction
            tpid = Get_Parameter(line, "tpid=", 10)
            predictions.append(tpid)

        if (( exit != None ) and valid ):
            (o3vtc_cost, iio_cost) = exit
            #print 'TRACE EXIT', exit

            if ( o3vtc_cost <= iio_cost ):
                # Breaks even
                be_cost += o3vtc_cost
                be_count += 1
                #print 'BREAKS EVEN'

            total_o3vtc_cost += o3vtc_cost
            total_iio_cost += iio_cost
            #print 'VT %x ET %s COSTS O3VTC %u IIO %u' % (
            #        vt_pc, et, o3vtc_cost, iio_cost)
            #print 'TOTAL COSTS O3VTC %u IIO %u' % (total_o3vtc_cost, 
            #                        total_iio_cost)
            valid = valid_next


        vt_pc = nvt_pc

    try:
        i = float(be_cost) / float(total_o3vtc_cost)
    except ZeroDivisionError:
        i = 0.0

    #print 'SRL END'
    #print '"iio inst_cost"', total_iio_cost
    #print '"o3vtc_cost"', total_o3vtc_cost

    return { "exit_count" : exit_count, 
            "trace_count" : trace_count,
            "be_proportion" : i,
            "be_count" : be_count,
            "o3vtc_cost" : total_o3vtc_cost,
            "inst_cost" : total_iio_cost,
            "iio_cost" : total_iio_cost,
            }

def Cleanup(name):
    for cleanup in [ '_1.txt', '_2.txt', '_run_00', '_run_01', '_run.txt' ]:
        try:
            os.unlink(name + cleanup)
        except:
            pass

def Test_Program(name):
    main_rc = Run_VT(name=name, capture_timings=False, execute_program=True)

    Crc(os.system("tail -n 100 %s_1.txt > %s_result2.txt" % (name, name)), 
                "tail") 

    cp_ok = Check_Path(name) 
    ok = cp_ok and ( main_rc == 0 )

    if ( not ok ):
        print 'Test_Program - Not OK'
        print 'Program', name
        print 'Check_Path', cp_ok
        print 'main_rc', main_rc
        return False

    #Cleanup(name)
    return True

def Main(timing_code):

    try:
        results = pickle.load(file("results_data"))
    except:
        results = dict()
   
    record = { "elf_path" : os.path.join(BENCH_PATH, "cnt"),
                "base_name" : "cnt",
            }
    program_list = [ record ]
    #program_list = program_list[ -2:-1 ]
    #print [ r[ 'base_name' ] for r in program_list ]
    #return 0
    vt_mode = (type(timing_code) == int)

    for record in program_list:
        elf = record[ 'elf_path' ]
        base_name = record[ 'base_name' ]
        #if ( base_name.find('edn') < 0 ):
        #    continue
        print base_name
        shutil.copy(elf, '%s.elf' % base_name)
        timings = Make_Timings(base_name, vt_mode)
        if ( timings == None ):
            return 1

        for keyname in timings.iterkeys():
            tc = timing_code
            results[ (base_name, keyname, tc) ] = timings[ keyname ]
      
        print timings

        pickle.dump(results, file("results_data", 'wb'))

    if ( vt_mode ):
        for record in program_list:
            base_name = record[ 'base_name' ]
            #if ( base_name.find('edn') < 0 ):
            #    continue
            if not Test_Program(base_name):
                return 1

    return 0


if ( __name__ == "__main__" ):
    print 'warning: this should only be called from build-staticN'
    assert len(sys.argv) == 2
    sys.exit(Main(sys.argv[ 1 ]))


