
import pickle

class Query:
    def __init__(self):
        self.tasks = pickle.load(file("tasks.dat"))
        #try:
        #    self.iio = pickle.load(file("../iio-store3.dat"))
        #except IOError:
        self.iio = dict()
        self.keys = self.tasks.keys()
        self.keys.sort()
        platforms = set()
        benchmarks = set()
        optimisations = set()
        lengths = set()

        for (name, length, experimental) in self.keys:
            (platform, benchmark, optimisation) = name.split('-')
            platforms.add(platform)
            benchmarks.add(benchmark)
            optimisations.add(optimisation)
            lengths.add(length)

        self.platforms = list(platforms)
        self.platforms.sort()
        self.benchmarks = list(benchmarks)
        self.benchmarks.sort()
        self.optimisations = list(optimisations)
        self.optimisations.sort()
        self.lengths = list(lengths)
        self.lengths.sort()

    def Get_Cross_Product(self, args):
        assert len(args) != 0
        assert type(args) in (list, tuple)
        assert type(args[ 0 ]) in (list, tuple)
        if len(args) == 1:
            return [ [x] for x in args[ 0 ] ]

        out = []
        for head in args[ 0 ]:
            for tail in self.Get_Cross_Product(args[ 1: ]):
                out.append([head] + tail)
                
        return out

    def Make_Tuple(self, in_list):
        out_list = []
        for key in in_list:
            if type(key) in (list, tuple):
                data = ','.join([ str(x) for x in key ])
                out_list.append((key, data))
            else:
                assert type(key) == str
                out_list.append((key, key))

        return out_list
                
    def Get_Row_Score(self, last_x_key, y_key):
        return self.Get_Cell(last_x_key, y_key, True)
   
    def Get_Cell(self, x_key, y_key, raw):
        return None

    def Get_Rows(self):
        return []

    def Get_Columns(self):
        return []

    def Get_Header(self):
        out = []
        out.append(r"\hline" + "\n")
        columns = self.Get_Columns()
        for (col_key, col_name) in columns:
            out.append(" &")
            out.append(col_name)
        out.append("\n")
        return ''.join(out)

    def Get_Footer(self):
        return r"\end{tabular}" + "\n\n"

    def Get_Table(self, latex=True, word="%22s", fout=None, 
            ignore_invalid=True):
        out = []
        if latex:
            hline = r"\hline" + "\n"
            endline = r"\\" + "\n"
            separator = "&"
            out.append(self.Get_Header())
            def Preproc(text):
                return (text.replace("janne_complex", "janne").
                            replace('%', '\\%').replace('_', '\\_'))

        else:
            hline = endline = "\n"
            separator = ","
            def Preproc(text):
                return text.replace('%', '')


        rows = self.Get_Rows()
        columns = self.Get_Columns()

        def GRS((row_key, row_name)):
            (col_key, col_name) = columns[ -1 ]
            return self.Get_Row_Score(col_key, row_key)

        rows.sort(key=GRS)

        out.append(hline)
        any_valid = False

        for (row_key, row_name) in rows:
            reset_to = len(out)
            valid = True
            out.append(word % Preproc(row_name))

            for (col_key, col_name) in columns:
                v = self.Get_Cell(col_key, row_key, not latex)
                if v == None:
                    valid = False
                    v = ""

                out.append(" ")
                out.append(separator)
                out.append(word % Preproc(str(v)))

            out.append(endline)
                
            if ignore_invalid and (not valid):
                out = out[ : reset_to ]

            any_valid = any_valid or valid

        assert any_valid
        out.append(hline)
        if latex:
            out.append(self.Get_Footer())
    
        out.append("\n")
        out = ''.join(out)
        if fout != None:
            file(fout, "wt").write(out)
        return out

    def Lookup(self, benchmark, platform=None, 
                    optimisation="3", length=None,
                    experimental=False):
        if length == None:
            length = self.lengths[ -1 ]
        if platform == None:
            platform = self.platforms[ -1 ]
        name = '%s-%s-%s' % (platform, benchmark, optimisation)
        d = self.tasks.get((name, length, experimental), None)
        if type(d) != dict:
            d = dict()
        if d.has_key("iio") and self.iio.has_key(name):
            d[ "iio" ] = self.iio[ name ]
        return d

    def Get_Percent(self, pc):
        return '%1.0f%%' % (100.0 * pc)

class ARMvsAlpha(Query):
    def Get_Columns(self):
        return self.Make_Tuple([
                    ("iio", "alpha"), ("iio", "arm"),
                    ("o3vtc", "alpha"), ("o3vtc", "arm"),
                    ("o3vtc", "diff"),
                    ("ncb", "alpha"), ("ncb", "arm"),
                    ])

    def Get_Row_Score(self, last_x_key, y_key):
        return self.Get_Cell(("o3vtc", "diff"), y_key, True)
   
    def Get_Header(self):
        return r"""
\begin{tabular}{|c|c|c|c|c|c|c|c|}
\hline
            & \multicolumn{2}{c|}{}
                & \multicolumn{2}{c|}{}
                & \tsize O3+VTC
                & \multicolumn{2}{c|}{\tsize Number of} \\
\tsize Benchmark   & \multicolumn{2}{c|}{\tsize IIO WCET} 
                & \multicolumn{2}{c|}{\tsize O3+VTC WCET} 
                & \tsize WCET: 
                & \multicolumn{2}{c|}{\tsize Cond. Branches} \\
            & \tsize Alpha & \tsize ARM & \tsize Alpha & \tsize ARM 
                & \tsize Alpha~$\div$~ARM 
                & \tsize Alpha & \tsize ARM \\                    
"""

    def Get_Rows(self):
        return [ (key, r"{\ccode\tsize  %s}" % key) for key in self.benchmarks ]
   
    def Get_Cell(self, x_key, y_key, raw):
        (field, platform) = x_key
        length = self.lengths[ -1 ]

        if platform != "diff":
            data = self.Lookup(platform=platform, 
                        benchmark=y_key,
                        length=length)
            out = data.get(field, None) 
            if out == None:
                return None
            elif raw:
                return out
            else:
                return r'\tsize %u' % out
        else:
            data = self.Lookup(platform="alpha", 
                        benchmark=y_key,
                        length=length)
            alpha = data.get(field, 0) 
            data = self.Lookup(platform="arm", 
                        benchmark=y_key,
                        length=length)
            arm = data.get(field, 0) 
            if 0 in (alpha, arm):
                return None
            
            out = float(alpha) / float(arm)
            if raw:
                return out
            else:
                return r'\tsize %1.3f' % out

class PlatQuery(Query):
    def __init__(self, my_platform="alpha"):
        Query.__init__(self)
        self.my_platform = my_platform

    def Get_Header(self):
        return r"""
\begin{tabular}{|c|c|c|c|c|c|c|}
\hline
            &       & \multicolumn{4}{c|}{{\tsize Improvement factor $f$}} \\
\tsize Benchmark   & \tsize IIO   & \multicolumn{4}{c|}{{\tsize with max. length}} \\
            & \tsize WCET  
""" + ''.join([ r'& \tsize %u ' % x for x in [ 1, 4, 8, 16 ] ]) + r"\\" + "\n"

    def Get_Columns(self):
        return self.Make_Tuple(["iio"]) + [ (x, str(x)) for x in 1, 4, 8, 16 ]

    def Get_Rows(self):
        return [ (key, r"{\ccode\tsize %s}" % key) for key in self.benchmarks ]
   
    def Get_Cell(self, x_key, y_key, raw):
        length = x_key
        benchmark = y_key
        platform = self.my_platform

        special = None
        if length == "iio":
            special = "iio"
            length = self.lengths[ 0 ] 
        elif length == 1:
            special = "triv"
            length = self.lengths[ 0 ] 

        data = self.Lookup(platform=platform, 
                    benchmark=benchmark,
                    length=length)

        iio = data.get("iio", 0)
        o3vtc = data.get("o3vtc", 0)

        if 0 in (iio, o3vtc):
            return None

        if special != None:
            x = data.get(special, None)
            if special == "iio":
                if raw:
                    return x
                else:
                    return r'\tsize %u' % x
            else:
                o3vtc = x

        pc = float(iio) / float(o3vtc)
        if raw:
            return -pc
        else:
            return r'\tsize %1.2f' % pc



def Main():
    PlatQuery("alpha").Get_Table(fout="article/article-alphatbl.tex", ignore_invalid=True)
    PlatQuery("arm").Get_Table(fout="article/article-armtbl.tex", ignore_invalid=True)
    ARMvsAlpha().Get_Table(fout="article/article-vstbl.tex", ignore_invalid=True)

if __name__ == "__main__":
    Main()


